using Adaptation.Eaf.Management.ConfigurationData.CellAutomation;
using Adaptation.FileHandlers.json.WorkItems;
using Adaptation.Ifx.Eaf.EquipmentConnector.File.Configuration;
using Adaptation.Shared;
using Adaptation.Shared.Duplicator;
using Adaptation.Shared.Methods;
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.IO;
using System.Linq;
using System.Text.Json;

namespace Adaptation.FileHandlers.Markdown;

public class FileRead : Shared.FileRead, IFileRead
{

    private readonly string _URL;
    private readonly ReadOnlyCollection<string> _WorkItemTypes;

    public FileRead(ISMTP smtp, Dictionary<string, string> fileParameter, string cellInstanceName, int? connectionCount, string cellInstanceConnectionName, FileConnectorConfiguration fileConnectorConfiguration, string equipmentTypeName, string parameterizedModelObjectDefinitionType, IList<ModelObjectParameterDefinition> modelObjectParameters, string equipmentDictionaryName, Dictionary<string, List<long>> dummyRuns, Dictionary<long, List<string>> staticRuns, bool useCyclicalForDescription, bool isEAFHosted) :
        base(new Description(), false, smtp, fileParameter, cellInstanceName, connectionCount, cellInstanceConnectionName, fileConnectorConfiguration, equipmentTypeName, parameterizedModelObjectDefinitionType, modelObjectParameters, equipmentDictionaryName, dummyRuns, staticRuns, useCyclicalForDescription, isEAFHosted: connectionCount is null)
    {
        _MinFileLength = 10;
        _NullData = string.Empty;
        _Logistics = new(this);
        if (_FileParameter is null)
            throw new Exception(cellInstanceConnectionName);
        if (_ModelObjectParameterDefinitions is null)
            throw new Exception(cellInstanceConnectionName);
        if (!_IsDuplicator)
            throw new Exception(cellInstanceConnectionName);
        string cellInstanceNamed = string.Concat("CellInstance.", _EquipmentType);
        _URL = GetPropertyValue(cellInstanceConnectionName, modelObjectParameters, $"{cellInstanceNamed}.URL");
        string workItemTypes = GetPropertyValue(cellInstanceConnectionName, modelObjectParameters, $"{cellInstanceNamed}.WorkItemTypes");
        _WorkItemTypes = new(workItemTypes.Split('|'));
        NestExistingFiles(_FileConnectorConfiguration);
    }

    void IFileRead.Move(Tuple<string, Test[], JsonElement[], List<FileInfo>> extractResults, Exception exception) => Move(extractResults);

    void IFileRead.WaitForThread() => WaitForThread(thread: null, threadExceptions: null);

    string IFileRead.GetEventDescription()
    {
        string result = _Description.GetEventDescription();
        return result;
    }

    List<string> IFileRead.GetHeaderNames()
    {
        List<string> results = _Description.GetHeaderNames();
        return results;
    }

    string[] IFileRead.Move(Tuple<string, Test[], JsonElement[], List<FileInfo>> extractResults, string to, string from, string resolvedFileLocation, Exception exception)
    {
        string[] results = Move(extractResults, to, from, resolvedFileLocation, exception);
        return results;
    }

    JsonProperty[] IFileRead.GetDefault()
    {
        JsonProperty[] results = _Description.GetDefault(this, _Logistics);
        return results;
    }

    Dictionary<string, string> IFileRead.GetDisplayNamesJsonElement()
    {
        Dictionary<string, string> results = _Description.GetDisplayNamesJsonElement(this);
        return results;
    }

    List<IDescription> IFileRead.GetDescriptions(IFileRead fileRead, List<Test> tests, IProcessData processData)
    {
        List<IDescription> results = _Description.GetDescriptions(fileRead, _Logistics, tests, processData);
        return results;
    }

    Tuple<string, Test[], JsonElement[], List<FileInfo>> IFileRead.GetExtractResult(string reportFullPath, string eventName)
    {
        Tuple<string, Test[], JsonElement[], List<FileInfo>> results;
        if (string.IsNullOrEmpty(eventName))
            throw new Exception();
        _ReportFullPath = reportFullPath;
        DateTime dateTime = DateTime.Now;
        results = GetExtractResult(reportFullPath, dateTime);
        if (results.Item3 is null)
            results = new Tuple<string, Test[], JsonElement[], List<FileInfo>>(results.Item1, Array.Empty<Test>(), JsonSerializer.Deserialize<JsonElement[]>("[]"), results.Item4);
        if (results.Item3.Length > 0 && _IsEAFHosted)
            WritePDSF(this, results.Item3);
        UpdateLastTicksDuration(DateTime.Now.Ticks - dateTime.Ticks);
        return results;
    }

    Tuple<string, Test[], JsonElement[], List<FileInfo>> IFileRead.ReExtract()
    {
        Tuple<string, Test[], JsonElement[], List<FileInfo>> results;
        List<string> headerNames = _Description.GetHeaderNames();
        Dictionary<string, string> keyValuePairs = _Description.GetDisplayNamesJsonElement(this);
        results = ReExtract(this, headerNames, keyValuePairs);
        return results;
    }

#nullable enable

    private static ReadOnlyDictionary<int, WorkItem> GetWorkItems(ReadOnlyCollection<WorkItem> workItems)
    {
        Dictionary<int, WorkItem> results = new();
        foreach (WorkItem workItem in workItems)
            results.Add(workItem.Id, workItem);
        return new(results);
    }

    private static int? GetIdFromUrlIfChild(Relation relation)
    {
        int? result;
        string[] segments = relation?.Attributes is null || relation.Attributes.Name != "Child" ? Array.Empty<string>() : relation.URL.Split('/');
        if (segments.Length < 2)
            result = null;
        else
        {
            if (!int.TryParse(segments[segments.Length - 1], out int id))
                result = null;
            else
                result = id;
        }
        return result;
    }

    private static Dictionary<int, Record> GetKeyValuePairs(ReadOnlyDictionary<int, WorkItem> workItems, WorkItem workItem)
    {
        Dictionary<int, Record> results = new();
        int? childId;
        WorkItem? childWorkItem;
        List<WorkItem> collection = new();
        Dictionary<int, Record> keyValuePairs;
        if (workItem.Relations is not null && workItem.Relations.Length > 0)
        {
            collection.Clear();
            foreach (Relation relation in workItem.Relations)
            {
                childId = GetIdFromUrlIfChild(relation);
                if (childId is null || !workItems.TryGetValue(childId.Value, out childWorkItem))
                    continue;
                collection.Add(childWorkItem);
            }
            collection = (from l in collection orderby l.State != "Closed", l.Id select l).ToList();
            foreach (WorkItem item in collection)
            {
                keyValuePairs = GetKeyValuePairs(workItems, item);
                results.Add(item.Id, new(item, new(keyValuePairs)));
            }
        }
        return results;
    }

    private static ReadOnlyDictionary<int, Record> GetWorkItemAndChildren(ReadOnlyDictionary<int, WorkItem> workItems)
    {
        Dictionary<int, Record> results = new();
        Dictionary<int, Record> keyValuePairs;
        foreach (KeyValuePair<int, WorkItem> keyValuePair in workItems)
        {
            // if (keyValuePair.Key != 119185)
            //     continue;
            keyValuePairs = GetKeyValuePairs(workItems, keyValuePair.Value);
            results.Add(keyValuePair.Key, new(keyValuePair.Value, new(keyValuePairs)));
        }
        return new(results);
    }

    private static string GetClosed(WorkItem workItem)
    {
        string result = workItem.State != "Closed" ? "[ ]" : "[x]";
        return result;
    }

    private static string GetLine(List<char> spaces, WorkItem workItem, KeyValuePair<int, Record> keyValuePair, bool condensed, bool sprintOnly) =>
        sprintOnly ? $"\t- [ ] {workItem.IterationPath}" :
            condensed ? $"{new string(spaces.Skip(1).ToArray())}- {GetClosed(workItem)} {keyValuePair.Key} - {workItem.Title}" :
            $"{new string(spaces.Skip(1).ToArray())}- {GetClosed(workItem)} {keyValuePair.Key} - {workItem.Title} ~~~ {workItem.AssignedTo} - {workItem.IterationPath.Replace('\\', '-')} - {workItem.CreatedDate} --- {workItem.ClosedDate}";

    private static void AppendLines(List<char> spaces, List<string> lines, Record record, bool condensed, bool sprintOnly)
    {
        string line;
        spaces.Add('\t');
        WorkItem workItem;
        foreach (KeyValuePair<int, Record> keyValuePair in record.Children)
        {
            workItem = keyValuePair.Value.WorkItem;
            line = GetLine(spaces, workItem, keyValuePair, condensed, sprintOnly).TrimEnd();
            lines.Add(line);
            AppendLines(spaces, lines, keyValuePair.Value, condensed, sprintOnly);
        }
        spaces.RemoveAt(0);
    }

    private static void AppendLines(string url, List<char> spaces, List<string> lines, ReadOnlyDictionary<int, Record> workItemAndChildren, string workItemType)
    {
        List<string> results = new();
        WorkItem workItem;
        string? maxIterationPath;
        List<string> distinct = new();
        foreach (KeyValuePair<int, Record> keyValuePair in workItemAndChildren)
        {
            workItem = keyValuePair.Value.WorkItem;
            // if (keyValuePair.Key != 109724)
            //     continue;
            if (workItem.WorkItemType != workItemType)
                continue;
            results.Add($"## {workItem.AssignedTo} - {workItem.Id} - {workItem.Title}");
            results.Add(string.Empty);
            results.Add($"- [{workItem.Id}]({url}{workItem.Id})");
            if (keyValuePair.Value.Children.Count == 0)
                results.Add(string.Empty);
            else
            {
                AppendLines(spaces, results, keyValuePair.Value, condensed: true, sprintOnly: false);
                results.Add(string.Empty);
                distinct.Clear();
                AppendLines(spaces, distinct, keyValuePair.Value, condensed: false, sprintOnly: true);
                if (distinct.Count > 1)
                {
                    results.Add($"## Distinct Iteration Path(s) - {workItem.WorkItemType} - {workItem.AssignedTo} - {workItem.Id} - {workItem.Title} - {workItem.IterationPath}");
                    results.Add(string.Empty);
                    results.Add($"- [{workItem.Id}]({url}{workItem.Id})");
                    distinct.Sort();
                    distinct = (from l in distinct select l.Trim()).Distinct().ToList();
                    results.AddRange(distinct);
                    results.Add(string.Empty);
                    maxIterationPath = distinct.Max();
                    if (!string.IsNullOrEmpty(maxIterationPath) && maxIterationPath.Contains("] ") && maxIterationPath.Split(']')[1].Trim() != workItem.IterationPath)
                    {
                        results.Add($"### Sync to Distinct Max Iteration Path => {maxIterationPath} - {workItem.Id} - {workItem.Title}");
                        results.Add(string.Empty);
                    }
                }
                results.Add($"## Extended - {workItem.Id} - {workItem.Title}");
                results.Add(string.Empty);
                AppendLines(spaces, results, keyValuePair.Value, condensed: false, sprintOnly: false);
                results.Add(string.Empty);
            }
            lines.AddRange(results);
            results.Clear();
        }
    }

    private static void WriteFiles(string destinationDirectory, string fileName, ReadOnlyCollection<string> lines, ReadOnlyCollection<WorkItem> workItems)
    {
        string text = string.Join(Environment.NewLine, lines);
        string markdownFile = Path.Combine(destinationDirectory, $"{fileName}.md");
        string textOld = !File.Exists(markdownFile) ? string.Empty : File.ReadAllText(markdownFile);
        if (text != textOld)
            File.WriteAllText(markdownFile, text);
        string html = CommonMark.CommonMarkConverter.Convert(text).Replace("<a href", "<a target='_blank' href");
        string htmlFile = Path.Combine(destinationDirectory, $"{fileName}.html");
        string htmlOld = !File.Exists(htmlFile) ? string.Empty : File.ReadAllText(htmlFile);
        if (html != htmlOld)
            File.WriteAllText(htmlFile, html);
        string json = JsonSerializer.Serialize(workItems.ToArray(), new JsonSerializerOptions() { WriteIndented = true });
        string jsonFile = Path.Combine(destinationDirectory, $"{fileName}.json");
        string jsonOld = !File.Exists(jsonFile) ? string.Empty : File.ReadAllText(jsonFile);
        if (json != jsonOld)
            File.WriteAllText(jsonFile, json);
    }

    private static ReadOnlyCollection<WorkItem> FilterChildren(Record record, ReadOnlyCollection<string> workItemTypes)
    {
        List<WorkItem> results = new();
        WorkItem workItem;
        foreach (KeyValuePair<int, Record> keyValuePair in record.Children)
        {
            workItem = keyValuePair.Value.WorkItem;
            if (!workItemTypes.Contains(workItem.WorkItemType))
                continue;
            results.Add(workItem);
        }
        return new(results);
    }

    private static string? GetMaxIterationPath(ReadOnlyCollection<WorkItem> workItems)
    {
        string? result;
        List<string> results = new();
        foreach (WorkItem workItem in workItems)
        {
            if (results.Contains(workItem.IterationPath))
                continue;
            results.Add(workItem.IterationPath);
        }
        result = results.Count == 0 ? null : results.Max();
        return result;
    }

    private static ReadOnlyCollection<WorkItem> FeatureCheckIterationPath122508(string url, List<string> lines, ReadOnlyCollection<string> workItemTypes, ReadOnlyDictionary<int, Record> workItemAndChildren, string workItemType)
    {
        List<WorkItem> collection = new();
        WorkItem workItem;
        string? maxIterationPath;
        List<string> results = new();
        ReadOnlyCollection<WorkItem> childrenWorkItems;
        foreach (KeyValuePair<int, Record> keyValuePair in workItemAndChildren)
        {
            workItem = keyValuePair.Value.WorkItem;
            if (workItem.WorkItemType != workItemType)
                continue;
            results.Clear();
            if (keyValuePair.Value.Children.Count == 0)
                continue;
            childrenWorkItems = FilterChildren(keyValuePair.Value, workItemTypes);
            maxIterationPath = GetMaxIterationPath(childrenWorkItems);
            if (string.IsNullOrEmpty(maxIterationPath) || workItem.IterationPath == maxIterationPath)
                continue;
            results.Add($"## {workItem.AssignedTo} - {workItem.Id} - {workItem.Title}");
            results.Add(string.Empty);
            results.Add($"- [ ] [{workItem.Id}]({url}{workItem.Id}) => {workItem.IterationPath} != {maxIterationPath}");
            results.Add(string.Empty);
            lines.AddRange(results);
            collection.Add(WorkItem.Get(workItem, $"IterationPath:{workItem.Id};{workItem.IterationPath} != {maxIterationPath}"));
            collection.Add(workItem);
        }
        return new(collection);
    }

    private static ReadOnlyCollection<WorkItem> GetWorkItemsNotMatching(string tags, ReadOnlyCollection<WorkItem> workItems)
    {
        List<WorkItem> results = new();
        string[] segments;
        string[] parentTags = tags.Split(';');
        foreach (WorkItem workItem in workItems)
        {
            segments = tags.Split(';');
            if (segments.Length > 0 && parentTags.Any(l => segments.Contains(l)))
                continue;
            results.Add(workItem);
        }
        return new(results);
    }

    private static ReadOnlyCollection<WorkItem> FeatureCheckTag122514(string url, List<string> lines, ReadOnlyCollection<string> workItemTypes, ReadOnlyDictionary<int, Record> workItemAndChildren, string workItemType)
    {
        List<WorkItem> collection = new();
        WorkItem workItem;
        List<string> results = new();
        List<string> violations = new();
        ReadOnlyCollection<WorkItem> childrenWorkItems;
        ReadOnlyCollection<WorkItem> workItemsNotMatching;
        foreach (KeyValuePair<int, Record> keyValuePair in workItemAndChildren)
        {
            workItem = keyValuePair.Value.WorkItem;
            if (workItem.WorkItemType != workItemType)
                continue;
            results.Clear();
            violations.Clear();
            if (keyValuePair.Value.Children.Count == 0)
                continue;
            if (string.IsNullOrEmpty(workItem.Tags))
                workItemsNotMatching = new(new WorkItem[] { workItem });
            else
            {
                childrenWorkItems = FilterChildren(keyValuePair.Value, workItemTypes);
                workItemsNotMatching = GetWorkItemsNotMatching(workItem.Tags, childrenWorkItems);
                if (!string.IsNullOrEmpty(workItem.Tags) && workItemsNotMatching.Count == 0)
                    continue;
            }
            results.Add($"## {workItem.AssignedTo} - {workItem.Id} - {workItem.Title}");
            results.Add(string.Empty);
            foreach (WorkItem item in workItemsNotMatching)
                results.Add($"- [ ] [{item}]({url}{item}) {nameof(workItem.Tags)} != {workItem.Tags}");
            results.Add(string.Empty);
            lines.AddRange(results);
            violations.Add($"Tag:{workItem.Tags};");
            foreach (WorkItem item in workItemsNotMatching)
                violations.Add($"{item.Id}:{item.Tags};");
            collection.Add(WorkItem.Get(workItem, string.Join(" ", violations)));
        }
        return new(collection);
    }

    private static ReadOnlyCollection<WorkItem> GetWorkItemsNotMatching(int? priority, ReadOnlyCollection<WorkItem> workItems)
    {
        List<WorkItem> results = new();
        foreach (WorkItem workItem in workItems)
        {
            if (workItem.Priority == priority)
                continue;
            results.Add(workItem);
        }
        return new(results);
    }

    private static ReadOnlyCollection<WorkItem> FeatureCheckPriority126169(string url, List<string> lines, ReadOnlyCollection<string> workItemTypes, ReadOnlyDictionary<int, Record> workItemAndChildren, string workItemType)
    {
        List<WorkItem> collection = new();
        WorkItem workItem;
        List<string> results = new();
        List<string> violations = new();
        ReadOnlyCollection<WorkItem> childrenWorkItems;
        ReadOnlyCollection<WorkItem> workItemsNotMatching;
        foreach (KeyValuePair<int, Record> keyValuePair in workItemAndChildren)
        {
            workItem = keyValuePair.Value.WorkItem;
            if (workItem.WorkItemType != workItemType)
                continue;
            results.Clear();
            violations.Clear();
            if (keyValuePair.Value.Children.Count == 0)
                continue;
            childrenWorkItems = FilterChildren(keyValuePair.Value, workItemTypes);
            workItemsNotMatching = GetWorkItemsNotMatching(workItem.Priority, childrenWorkItems);
            if (workItemsNotMatching.Count == 0)
                continue;
            results.Add($"## {workItem.AssignedTo} - {workItem.Id} - {workItem.Title}");
            results.Add(string.Empty);
            results.Add($"- [{workItem.Id}]({url}{workItem.Id})");
            foreach (WorkItem item in workItemsNotMatching)
                results.Add($"- [ ] [{item}]({url}{item}) {nameof(workItem.Priority)} != {workItem.Priority}");
            results.Add(string.Empty);
            lines.AddRange(results);
            violations.Add($"Priority:{workItem.Priority};");
            foreach (WorkItem item in workItemsNotMatching)
                violations.Add($"{item.Id}:{item.Priority};");
            collection.Add(WorkItem.Get(workItem, string.Join(" ", violations)));
        }
        return new(collection);
    }

    private static void WriteFiles(FileConnectorConfiguration fileConnectorConfiguration, string url, ReadOnlyCollection<string> workItemTypes, ReadOnlyCollection<WorkItem> workItems)
    {
        string json;
        List<char> spaces = new();
        List<string> lines = new();
        ReadOnlyCollection<WorkItem> results;
        ReadOnlyDictionary<int, WorkItem> collection = GetWorkItems(workItems);
        ReadOnlyDictionary<int, Record> keyValuePairs = GetWorkItemAndChildren(collection);
        ReadOnlyCollection<string> bugUserStoryWorkItemTypes = new(new string[] { "Bug", "User Story" });
        ReadOnlyCollection<string> bugUserStoryTaskWorkItemTypes = new(new string[] { "Bug", "User Story", "Task" });
        if (!Directory.Exists(fileConnectorConfiguration.TargetFileLocation))
            _ = Directory.CreateDirectory(fileConnectorConfiguration.TargetFileLocation);
        if (keyValuePairs.Count == -1)
        {
            json = JsonSerializer.Serialize(keyValuePairs, new JsonSerializerOptions() { WriteIndented = true });
            File.WriteAllText(Path.Combine(fileConnectorConfiguration.TargetFileLocation, ".json"), json);
        }
        foreach (string workItemType in workItemTypes)
        {
            lines.Clear();
            lines.Add($"# {workItemType}");
            lines.Add(string.Empty);
            AppendLines(url, spaces, lines, keyValuePairs, workItemType);
            results = new(Array.Empty<WorkItem>());
            WriteFiles(fileConnectorConfiguration.TargetFileLocation, workItemType, new(lines), results);
        }
        {
            lines.Clear();
            string workItemType = "Feature";
            lines.Add($"# {nameof(FeatureCheckIterationPath122508)}");
            lines.Add(string.Empty);
            results = FeatureCheckIterationPath122508(url, lines, bugUserStoryTaskWorkItemTypes, keyValuePairs, workItemType);
            WriteFiles(fileConnectorConfiguration.TargetFileLocation, "check-122508", new(lines), results);
        }
        {
            lines.Clear();
            string workItemType = "Feature";
            lines.Add($"# {nameof(FeatureCheckTag122514)}");
            lines.Add(string.Empty);
            results = FeatureCheckTag122514(url, lines, bugUserStoryWorkItemTypes, keyValuePairs, workItemType);
            WriteFiles(fileConnectorConfiguration.TargetFileLocation, "check-122514", new(lines), results);
        }
        {
            lines.Clear();
            string workItemType = "Feature";
            lines.Add($"# {nameof(FeatureCheckPriority126169)}");
            lines.Add(string.Empty);
            results = FeatureCheckPriority126169(url, lines, bugUserStoryWorkItemTypes, keyValuePairs, workItemType);
            WriteFiles(fileConnectorConfiguration.TargetFileLocation, "check-126169", new(lines), results);
        }
    }

#pragma warning disable IDE0060
    private void WriteFiles(string reportFullPath, DateTime dateTime)
#pragma warning restore IDE0060
    {
        string json = File.ReadAllText(reportFullPath);
        WorkItem[]? workItems = JsonSerializer.Deserialize<WorkItem[]>(json);
        if (workItems is null)
            throw new Exception(nameof(workItems));
        if (workItems.Length > 0)
            WriteFiles(_FileConnectorConfiguration, _URL, _WorkItemTypes, new(workItems));
    }

    private Tuple<string, Test[], JsonElement[], List<FileInfo>> GetExtractResult(string reportFullPath, DateTime dateTime)
    {
        Tuple<string, Test[], JsonElement[], List<FileInfo>> results;
        _Logistics = new Logistics(reportFullPath, $"LOGISTICS_1{'\t'}A_JOBID={"BACKLOG"};A_MES_ENTITY={"BACKLOG"};");
        if (_IsEAFHosted)
            WriteFiles(reportFullPath, dateTime);
        results = new(_Logistics.Logistics1[0], Array.Empty<Test>(), Array.Empty<JsonElement>(), new List<FileInfo>());
        return results;
    }

}